package com.engidea.winjcom;

import com.engidea.winjcom.exception.UnsupportedCommOperationException;
import java.io.*;


/**
  WinJcom is a native interface to serial ports in java.
  Copyright 2007 by Damiano Bolla, Jcomm@engidea.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.
  This can be used with commercial products and you are not obliged 
  to share your work with anybody.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


public final class JcomPort extends SerialPort
  {
  private static final String classname="JcomPort.";
  private static final boolean debug = false;

  private final SerialOutputStream portOutStream = new SerialOutputStream();
  private final SerialInputStream  portInStream  = new SerialInputStream();

  private int    portFd = -1;
  private Thread monitorThread;
  private SerialPortEventListener portListener;


  /**
   * Constructor.
   * @param portName
   */
  public JcomPort( String portName)
    {
    super(portName);
    }


  private native int nativeOpen(String name) throws IOException;

  /**
   * This is an addition to the standard.
   * @throws IOException
   */
  public void open () throws IOException
    {
    portFd = nativeOpen(portName);

    // You need to set at least one event...
    nativeNotifyEnable(portFd,EV_RING,true);

    if (debug) report (classname+"Constructor: openFd="+portFd);
    
    monitorThread = new Thread(new MonitorThread());
    monitorThread.start();
    }

  private void report ( String msg )
    {
    System.out.println(msg);
    }

  public OutputStream getOutputStream()
    {
    return portOutStream;
    }

  public InputStream getInputStream()
    {
    return portInStream;
    }

  private native void nativeSetSerialPortParams (
    int portFd,
    int baudrate, 
    int dataBits, 
    int stopBits, 
    int parity) throws UnsupportedCommOperationException;


  public synchronized void setSerialPortParams(
    int baudrate, 
    int databits, 
    int stopbits, 
    int parity) throws UnsupportedCommOperationException
    {
    nativeSetSerialPortParams(portFd,baudrate,databits,stopbits,parity);
    }


  /**
   * Native params are all accessed using a single native call.
   * param type is defined here as a constant.
   * @param paramType
   * @return
   */
  private static final int DCB_BAUDRATE    = 1;   
  private static final int DCB_DATABITS    = 2;
  private static final int DCB_STOPBITS    = 3;
  private static final int DCB_PARITY      = 4;
  private static final int DCB_FLOWCONTROL = 5;

  private native int nativeGetPortParam (int portFd, int paramType);
  
  public int getBaudRate()
    {
    return nativeGetPortParam(portFd, DCB_BAUDRATE);
    }


  public int getDataBits()
    {
    return nativeGetPortParam(portFd, DCB_DATABITS);
    }

  public int getStopBits()
    {
    return nativeGetPortParam(portFd, DCB_STOPBITS);
    }

  public int getParity()
    {
    return nativeGetPortParam(portFd, DCB_PARITY);
    }

  public int getFlowControlMode()
    {
    return nativeGetPortParam(portFd, DCB_FLOWCONTROL);
    }


  /**
   * Note that what to tinker with is one of DCB_ defined in here
   * @param portFd
   * @param what    one of DCB_ defined in here
   * @param value   depends on what you want to set...
   */
  private native void nativeSetPortParam ( int portFd, int what, int value );

  /**
   * The param is one of FLOWCONTROL_ defined in SerialPort
   * @param flowcontrol
   */
  public void setFlowControlMode(int flowcontrol)
    {
    nativeSetPortParam(portFd,DCB_FLOWCONTROL, flowcontrol);
    }

  public void enableReceiveFraming(int f) throws UnsupportedCommOperationException
    {
    throw new UnsupportedCommOperationException("Unsupported");
    }

  public void disableReceiveFraming()
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public boolean isReceiveFramingEnabled()
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public int getReceiveFramingByte()
    {
    throw new IllegalArgumentException("Not implemented");
    }


  /**
   * ALl values are in milliseconds.
   * @param portFd
   * @param reInTi
   * @param reToTiMu
   * @param reToTiCo
   * @param wrToTiMu
   * @param wrToTiCo
   */
  private native void nativeSetCommTimeouts ( 
    int portFd, 
    int reInTi, 
    int reToTiMu, 
    int reToTiCo,
    int wrToTiMu,
    int wrToTiCo );

  public void disableReceiveTimeout()
    {
    nativeSetCommTimeouts(portFd, 0,0,0, 0,0);
    }

  public void enableReceiveTimeout(int timeoutMilli)
    {
    // This sets a timeout between chars as much as a total timeout
    // it also sets the SAME timeout for both read and write.
    nativeSetCommTimeouts(portFd, timeoutMilli,0,timeoutMilli, 0,timeoutMilli);
    }

  private native int nativeGetReceiveTotalTimeoutConstant ( int portFd );
  
  public boolean isReceiveTimeoutEnabled()
    {
    return nativeGetReceiveTotalTimeoutConstant ( portFd ) != 0;
    }

  public int getReceiveTimeout()
    {    
    return nativeGetReceiveTotalTimeoutConstant ( portFd );
    }

  public void enableReceiveThreshold(int thresh)
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public void disableReceiveThreshold()
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public int getReceiveThreshold()
    {    
    throw new IllegalArgumentException("Not implemented");
    }

  public boolean isReceiveThresholdEnabled()
    {
    throw new IllegalArgumentException("Not implemented");
    }


  public void setInputBufferSize(int size)
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public int getInputBufferSize()
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public void setOutputBufferSize(int size)
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public int getOutputBufferSize()
    {
    throw new IllegalArgumentException("Not implemented");
    }

  public boolean isDTR()
    {
    throw new IllegalArgumentException("Not implemented");
    }


  /**
   * The params are taken directly from Win 
   * @param portFd
   * @param what
   */
  private native void nativeSendEscape ( int portFd, int what );
  
  private static final int ESC_SETRTS=3;  // As from windows !
  private static final int ESC_CLRRTS=4;  // As from windows !
  private static final int ESC_SETDTR=5;  // As from windows !
  private static final int ESC_CLRDTR=6;  // As from windows !
  private static final int ESC_SETBREAK=8;  // As from windows !
  private static final int ESC_CLRBREAk=9;  // As from windows !

  public void  setDTR(boolean state)
    {
    int what = state ? ESC_SETDTR : ESC_CLRDTR;
    
    nativeSendEscape (portFd, what);
    }

  public void  setRTS(boolean state)
    {
    int what = state ? ESC_SETRTS : ESC_CLRRTS;
    
    nativeSendEscape (portFd, what);
    }

  public  void sendBreak(int duration) throws InterruptedException
    {
    nativeSendEscape (portFd, ESC_SETBREAK);
    Thread.currentThread().sleep(duration);
    nativeSendEscape (portFd, ESC_CLRBREAk);
    }


  private static final int MS_CTS_ON=16;
  private static final int MS_DSR_ON=32;
  private static final int MS_RING_ON=64;
  private static final int MS_RLSD_ON=128;
  
  private native int nativeGetModemStatus ( int fd );

  public  boolean isCTS()
    {
    return (nativeGetModemStatus(portFd) & MS_CTS_ON) != 0;
    }

  public  boolean isDSR()
    {
    return (nativeGetModemStatus(portFd) & MS_DSR_ON) != 0;
    }

  public  boolean isCD()
    {
    return (nativeGetModemStatus(portFd) & MS_RLSD_ON) != 0;
    }
    
  public  boolean isRI()
    {
    return (nativeGetModemStatus(portFd) & MS_RING_ON) != 0;
    }

  public  boolean isRTS()
    {
    throw new IllegalArgumentException("Not implemented");
    }

  /**
   * At the moment sets the listener, in the future I will have a list of them...
   * @param portListener
   */
  public void addEventListener(SerialPortEventListener portListener)
    {
    this.portListener = portListener;
    }

  /**
   *  Remove the serial port event listener
   */
  public void removeEventListener()
    {
    portListener = null;
    }


  private static final int EV_RXCHAR=0x0001;   // Any Character received
  private static final int EV_CTS=0x0008;      // CTS received
  private static final int EV_TXEMPTY=0x0004;  // Notify on output empty
  private static final int EV_DSR=0x0010;      // DSR changed state
  private static final int EV_BREAK=0x0040;    // BREAK received
  private static final int EV_ERR=0x0080;      // Line status error occurred
  private static final int EV_RING=0x0100;     // Ring signal detected
  private static final int EV_RLSD=0x0020;     // The RLSD (receive-line-signal-detect) signal changed state


  /**
   * Enable or disable notification.
   * You MUST call this one with dummy params at the beginning to setup things properly.
   * @param portFd
   * @param what    a value as defined in SetCommMask, keep them the same...
   * @param enable
   * @throws IOException
   */
  private native void nativeNotifyEnable ( int portFd, int what, boolean enable );

  public void notifyOnDataAvailable(boolean enable)   
    {   
    nativeNotifyEnable(portFd,EV_RXCHAR,enable);   
    }
    
  public void notifyOnOutputEmpty(boolean enable)     
    {   
    nativeNotifyEnable(portFd,EV_TXEMPTY,enable);  
    }

  public void notifyOnCTS(boolean enable)             
    {   
    nativeNotifyEnable(portFd,EV_CTS,enable);      
    }
    
  public void notifyOnDSR(boolean enable)             
    {   
    nativeNotifyEnable(portFd,EV_DSR,enable);      
    }
    
  public void notifyOnRingIndicator(boolean enable)   
    {   
    nativeNotifyEnable(portFd,EV_RING,enable);     
    }
    
  public void notifyOnCarrierDetect(boolean enable)   
    {   
    nativeNotifyEnable(portFd,EV_RLSD,enable);     
    }
    
  public void notifyOnOverrunError(boolean enable)    
    {   
    nativeNotifyEnable(portFd,EV_ERR,enable);      
    }
    
  public void notifyOnParityError(boolean enable)     
    {   
    nativeNotifyEnable(portFd,EV_ERR,enable);      
    }
    
  public void notifyOnFramingError(boolean enable)    
    {   
    nativeNotifyEnable(portFd,EV_ERR,enable);      
    }
  
  public void notifyOnBreakInterrupt(boolean enable)  
    {   
    nativeNotifyEnable(portFd,EV_BREAK,enable);    
    }



  private native void nativeClose(int portFd) throws IOException;

  public synchronized void close() 
    {
    try
      {
      nativeClose(portFd);
      }
    catch ( Exception exc )
      {
      exc.printStackTrace();
      }
      
    portFd = -1;
    // the event thread just dies since I have closed the data stream.
    }


  /**
   * Ideally the program MUST call close, but if it does not this is a chance to recover.
   * Of course you never know when GC is run...
   */
  public void finalize()
    {
    if (portFd < 0) return;
    
    report("finalize:calling close()");
  
    close();
    }


  private native int nativeWriteArray ( int portFd, byte[]buff, int offset, int len ) throws IOException;
  private native void nativePurgeComm ( int portFd ) throws IOException;

/**
 * NOTE: If you want to know how many bytes have been written before the timeout
 * you should use something different than OutputStream since that does not return
 * the number of bytes written !
 * You can make your own function that calls nativeWriteArray !
 */
public final class SerialOutputStream extends OutputStream
  {
  public void write(int val) throws IOException
    {
    byte []value = new byte[1];

    value[0] = (byte)val;
    
    write(value);
    }

  public void write(byte[] buff) throws IOException
    {
    write(buff, 0, buff.length);
    }

  public void write(byte[] buff, int offset, int len) throws IOException
    {
    int written = nativeWriteArray(portFd, buff, offset, len);
    
    if ( written != len )
      throw new IOException("Timeout len="+len+" written="+written);
    }

  public void flush() throws IOException
    {
    nativePurgeComm(portFd);
    }
  }


private static final int COMSTAT_cbInQue = 1;   // Number of bytes received but not yet read

private native int nativeClearCommError( int portFd, int what ) throws IOException;;

private native int nativeReadArray (int portFd, byte[] buff, int offset, int len) throws IOException;

/**
 * Note that the normal contrat for read is that it returns the whole bytes unless we are at EOF
 * and in that case it returns -1, so, if I have a timeout I should really throw an Exception
 * like the write part. BUT the native part allows you to find out what is happening
 * This is a slightly different contract from the normal read....
 */
public final class SerialInputStream extends InputStream
  {
  public synchronized int read() throws IOException
    {
    byte []risul = new byte[1];

    int letti = read(risul);

    if ( letti == 1 ) return risul[0] & 0x00FF;
    
    return letti;
    }

  public synchronized int read(byte[] buff) throws IOException
    {
    return read(buff, 0, buff.length );
    }

  public synchronized int read(byte[] buff, int offset, int len) throws IOException
    { 
    int letti = nativeReadArray(portFd, buff, offset, len );
    
    if ( letti != len )
      throw new IOException("Timeout len="+len+" letti="+letti);
    
    return letti;
    }

  public synchronized int available() throws IOException
    {   
    return nativeClearCommError(portFd, COMSTAT_cbInQue);
    }
  }



  
private native int nativeGetCommEvent (int portFd) throws IOException;

private class MonitorThread implements Runnable
    {
    private void waitForEvent () throws IOException
      {
      int eventMask = nativeGetCommEvent(portFd);

      if ( portListener == null ) return;
      
      if ( (eventMask & EV_CTS) != 0 ) 
        portListener.serialEvent(new SerialPortEvent(JcomPort.this, SerialPortEvent.CTS,false,isCTS()));
        
      if ( (eventMask & EV_RING) != 0 ) 
        portListener.serialEvent(new SerialPortEvent(JcomPort.this, SerialPortEvent.RI,false,isRI()));
        
      if ( (eventMask & EV_DSR) != 0 ) 
        portListener.serialEvent(new SerialPortEvent(JcomPort.this, SerialPortEvent.DSR,false,isDSR()));
        
      if ( (eventMask & EV_RLSD) != 0 ) 
        portListener.serialEvent(new SerialPortEvent(JcomPort.this, SerialPortEvent.CD,false,isCD()));

      if ( (eventMask & EV_TXEMPTY) != 0 ) 
        portListener.serialEvent(new SerialPortEvent(JcomPort.this, SerialPortEvent.OUTPUT_BUFFER_EMPTY,false,true));

      if ( (eventMask & EV_RXCHAR) != 0 ) 
        portListener.serialEvent(new SerialPortEvent(JcomPort.this, SerialPortEvent.DATA_AVAILABLE,false,true));
      }

    public void run()
      {
      if ( debug ) report("MonitorThread: port("+portName+") start");

      try
        {
        for(;;) waitForEvent();
        }
      catch ( Exception exc )
        {
        // Monitor thread exit when the port is closed
        // The result is an exception that I could catch or just ignore...
        if ( debug ) report("MonitorThread: port("+portName+") Exception="+exc);
        }

      if ( debug ) report("MonitorThread: port("+portName+") exit");
      }

    }


}
